import datetime
import logging
from typing import List, Tuple

from pydantic_db_backend.backend import Backend
from pydantic_db_backend_common.exceptions import RevisionConflict
from pydantic_db_backend_common.utils import utcnow

from eventix.exceptions import NoTaskFoundForUniqueKey
from eventix.pydantic.pagination import PaginationParametersModel
from eventix.pydantic.task import TaskModel, task_model_status

# post
# already exists
# still scheduled -> update
# not scheduled --> post new.
# scheduled -> update
#
#   update:
#       get task
#       scheck scheduled, if not post
#       update task
#
#       if worker grabs , version should conflict
#       if conflict: try again

log = logging.getLogger(__name__)


def task_post(task: TaskModel) -> TaskModel:
    is_unique = task.unique_key is not None

    if not is_unique:
        # not unique , just try to save. If exists, raise error
        # noinspection PyTypeChecker
        return Backend.post_instance(task)

    # has unique_key

    while True:

        # noinspection PyTypeChecker
        existing_tasks: List[TaskModel] = Backend.get_instances(TaskModel, 0, 10, {"unique_key": task.unique_key})
        next_scheduled_task = next(filter(lambda t: t.status in ("scheduled", "retry"), existing_tasks), None)

        if next_scheduled_task is None:
            # no existing ones that are only scheduled, we have to post
            # noinspection PyTypeChecker
            return Backend.post_instance(task)

        #   update:
        #       get task
        #       scheck scheduled, if not post
        #       update task
        #
        #       if worker grabs , version should conflict
        #       if conflict: try again

        next_scheduled_task.unique_update_from(task)
        try:
            updated_task = Backend.put_instance(next_scheduled_task)
            # noinspection PyTypeChecker
            log.debug(f"updated task {updated_task.uid}")
            # noinspection PyTypeChecker
            return updated_task  # update worked
        except RevisionConflict as e:
            continue  # try again.


def task_clean_expired_workers():
    params = dict(
        model=TaskModel,
        skip=0,
        limit=1,
        query_filter=dict(
            worker_expires={
                "$and": [
                    {"$ne": None},
                    {"$lt": utcnow()},
                ]
            }  # no worker assigned
        ),
        sort=[
            {"priority": "asc"},
            {"eta": "asc"}
        ]
    )

    while True:
        # noinspection PyTypeChecker
        existing_task: TaskModel | None = next(iter(Backend.get_instances(**params)), None)

        # repeat until we were able to take something or nothing is left.
        if existing_task is None:
            break

        existing_task.status = "scheduled"
        existing_task.worker_id = None
        existing_task.worker_expires = None

        try:
            Backend.put_instance(existing_task)
            log.info(f"Released task {existing_task.uid}")
            # noinspection PyTypeChecker
        except RevisionConflict as e:
            continue


def task_clean_expired_tasks():
    params = dict(
        model=TaskModel,
        skip=0,
        limit=100,
        query_filter=dict(
            expires={
                "$and": [
                    {"$ne": None},
                    {"$lt": utcnow()},
                ]
            }  # task expired
        ),
        # sort=[
        #     {"priority": "asc"},
        #     {"eta": "asc"}
        # ]
    )

    while True:
        # noinspection PyTypeChecker
        existing_uids = Backend.get_uids(**params)

        # repeat until we were able to take something or nothing is left.
        if len(existing_uids) == 0:
            break

        for uid in existing_uids:
            Backend.delete_uid(TaskModel, uid)
            log.info(f"Removed expired task {uid}")


def task_next_scheduled(worker_id: str, namespace: str, expires: int = 300) -> TaskModel | None:
    log.debug(f"[{worker_id}] Worker getting next scheduled task...")

    # looking up possible tasks in right order
    # take first one
    # try to set worker_id and expiration

    eta = utcnow().isoformat()  # eta has to be now or in the past

    query_filter = dict(
        namespace=namespace,  # namespace has to match
        worker_id=None,  # no worker assigned
        status={"$in": ["scheduled", "retry"]},
        eta={"$lte": eta}
    )
    sort = [
        {"priority": "asc"},
        {"eta": "asc"}
    ]

    while True:  # repeat until we were able to take something or nothing is left.

        # noinspection PyTypeChecker
        existing_task: TaskModel | None = next(iter(Backend.get_instances(
            TaskModel,
            0,
            1,
            query_filter=query_filter,
            sort=sort
        )), None)

        if existing_task is None:
            return None  # no task left

        existing_task.status = "processing"
        existing_task.worker_id = worker_id
        existing_task.worker_expires = utcnow() + datetime.timedelta(seconds=expires)
        log.debug(f"task_next_scheduled: existing task revision: {existing_task.revision}")
        try:
            # noinspection PyTypeChecker
            t: TaskModel = Backend.put_instance(existing_task)
            return t
        except RevisionConflict as e:
            continue


def tasks_by_status(
    status: task_model_status | None = None,
    namespace: str | None = None,
    pagination: PaginationParametersModel | None = None,
    max_results: bool | None = True
) -> Tuple[List[TaskModel], int]:
    query_filter = {}
    if status is not None:
        query_filter["status"] = {"$eq": status}
    if namespace is not None:
        query_filter["namespace"] = {"$eq": namespace}
    params = {
        "query_filter": query_filter,
        **pagination.dict(),
    }

    # noinspection PyTypeChecker
    tasks = Backend.get_instances(TaskModel, **params, max_results=max_results)
    if max_results is True:
        tasks, max_results = tasks
    else:
        max_results = 0

    tasks: List[TaskModel]
    return tasks, max_results


def task_by_unique_key(unique_key: str) -> TaskModel:
    # noinspection PyTypeChecker
    existing_tasks: List[TaskModel] = Backend.get_instances(TaskModel, 0, 10, {"unique_key": unique_key})
    found = next(filter(lambda t: t.status in ("scheduled", "retry"), existing_tasks), None)
    if found is None:
        raise NoTaskFoundForUniqueKey(unique_key=unique_key)

    return found



import importlib

import toml
from pathlib import Path


def get_version():
    try:
        path = Path(__file__).resolve().parents[1] / 'pyproject.toml'
        pyproject = toml.loads(open(str(path)).read())
        v = pyproject['tool']['poetry']['version']

    except Exception:
        try:
            v = importlib.metadata.version('eventix')
        except Exception:
            v = "0.0.0"
    return v


__version__ = get_version()



from eventix.functions.core import task
from eventix.functions.task import task_clean_expired_workers, task_clean_expired_tasks


@task(unique_key_generator=lambda: "cleanup_worker", store_result=False)
def task_cleanup_worker():
    task_clean_expired_workers()


@task(unique_key_generator=lambda: "cleanup_tasks", store_result=False)
def task_cleanup_results():
    task_clean_expired_tasks()



from collections import OrderedDict
from typing import Optional, List, Dict, Callable
from eventix import __version__

import click
import lsjsonclasses
import pydash
from lsrestclient import LsRestClient
from pydantic import BaseModel

from eventix.pydantic.task import TaskModel


class CLIContext(BaseModel):
    client: LsRestClient
    namespace: Optional[str]

    class Config:
        arbitrary_types_allowed = True


@click.group()
@click.version_option(__version__)
@click.option("-s", "--server", help="server url")
@click.option("-n", "--namespace", help="namespace")
@click.pass_context
def cli(ctx, server: str = None, namespace: str = None):
    ctx.obj = CLIContext(
        namespace=namespace,
        client=LsRestClient(base_url=server, name="eventix")
    )


@cli.group("get")
@click.pass_context
def cli_get(ctx):
    pass


@cli_get.command("tasks")
@click.option("--limit", "limit", default=10, type=int)
@click.option("--skip", "skip", default=0, type=int)
@click.option("--status", "status", default=None, type=str)
@click.pass_context
def cli_get_tasks(ctx, skip: int, limit: int, status: str | None = None):
    namespace = ctx.obj.namespace

    body = dict(
        skip=skip,
        limit=limit
    )

    body |= dict(
        namespace=namespace,
        status=status
    )
    # pprint(body)

    r = ctx.obj.client.put("/tasks/by_status", body=body)
    if r.status_code == 200:
        pv = r.json()
        max_result = pydash.default_to(pydash.get(pv, "max_results"), 0)

        start = skip + 1
        end = skip + limit if skip + limit < max_result else max_result
        pagination_result = f"{start}-{end} of {max_result}"

        tasks = [TaskModel.parse_obj(x) for x in pydash.get(pv, 'data', [])]

        if len(tasks) != 0:
            entries = [
                OrderedDict(
                    NAMESPACE=task.namespace,
                    UID=task.uid,
                    TASK=task.task,
                    STATUS=task.status,
                    IDENTIFIER=""
                ) for task in tasks
            ]
            print_table(entries, click.echo)

            click.echo(f"Results {pagination_result}")
        else:
            click.echo(f"No result")


@cli_get.command("task")
@click.option("--error-only", is_flag=True, default=False)
@click.argument("uid")
@click.pass_context
def cli_get_task(ctx, uid: str, error_only: bool):
    r = ctx.obj.client.get("/task/{uid}", params=dict(uid=uid))
    if r.status_code == 200:
        data = r.json()
        result = pydash.get(data, "result", None)
        if isinstance(result, dict):
            error_class = pydash.get(result, "error_class", None)
        else:
            error_class = None

        if not error_only:
            if error_class is not None:
                data['result'] = pydash.pick(result, "error_class")

            print_dict_to_json(data, click.echo)
        else:
            click.echo(f"ERROR_CLASS: {error_class}")
            if error_class == "LSoftException":
                error_dict = lsjsonclasses.LSoftJSONDecoder.loads(result['error_message'])
                error = pydash.get(error_dict, "detail.ERROR", None)
                traceback_str = pydash.get(error_dict, "detail.TRACEBACK", None)
                click.echo(f"ERROR:")
                click.echo(error)
                click.echo(f"TRACEBACK:")
                click.echo(traceback_str)
            else:
                print_dict_to_json(result, click.echo)

    else:
        click.echo(f"Error: {r.content}")


@cli.group("delete")
@click.pass_context
def cli_delete(ctx):
    pass


@cli_delete.command("task")
@click.argument("uid")
@click.pass_context
def cli_get_task(ctx, uid: str):
    r = ctx.obj.client.delete("/task/{uid}", params=dict(uid=uid))
    if r.status_code == 200:
        click.echo(f"Deleted task {uid}")
    else:
        click.echo(f"Error: {r.content}")


def print_dict_to_json(data: dict, print_func: Callable = print):
    j = lsjsonclasses.LSoftJSONEncoder.dumps(data, indent=2)
    print_func(j)


def print_table(tdata: List[Dict[str, str]], print_func: Callable = print):
    if len(tdata) == 0:
        return

    keys = list(tdata[0].keys())
    sizes = {}
    lines = []
    for key in keys:
        sizes[key] = max(pydash.concat([len(key)], [len(x[key]) for x in tdata]))

    tdata = pydash.concat([{k: k for k in keys}], tdata)
    for td in tdata:
        line = ""
        for key in keys:
            line += "{text:{width}} ".format(text=td[key], width=sizes[key])
        lines.append(line.strip())

    for line in lines:
        print_func(line)


if __name__ == '__main__':
    cli()



import logging
import os
import sys

from eventix.contexts import namespace_context, namespace_context_var, eventix_base_url_context
from eventix.exceptions import backend_exceptions
from eventix.functions.errors import raise_errors
from eventix.functions.eventix_client import EventixClient
from eventix.pydantic.task import TaskModel
import json
import datetime

log = logging.getLogger(__name__)


class LSoftJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()

    @classmethod
    def dumps(cls, obj, *args, **kwargs):
        return json.dumps(obj, *args, cls=cls, **kwargs)


class TaskScheduler(EventixClient):

    @classmethod
    def schedule(cls, task: TaskModel) -> TaskModel:
        if task.namespace is None:
            with namespace_context() as namespace:
                task.namespace = namespace
        log.debug(f"scheduling task: {task.task} namespace: {task.namespace} uid: {task.uid} eta: {task.eta} unique_key: {task.unique_key} priority: {task.priority}")
        return cls.task_post(task)

    @classmethod
    def task_get(cls, uid: str) -> TaskModel:
        r = cls.interface.get(f'/task/{uid}')
        with raise_errors(r, backend_exceptions):
            return TaskModel.parse_raw(r.content)

    @classmethod
    def task_post(cls, task: TaskModel) -> TaskModel:
        # TODO: use Pydantic v2 json
        a = LSoftJSONEncoder.dumps(task.model_dump())
        with eventix_base_url_context() as ctx_base_url:
            main_url = cls.interface.base_url
            if ctx_base_url is not None:
                cls.interface.base_url = ctx_base_url
            r = cls.interface.post('/task', data=a)
            cls.interface.base_url = main_url
        with raise_errors(r, backend_exceptions):
            return TaskModel.parse_raw(r.content)

    @classmethod
    def config(cls, config: dict):
        # Be aware that the namespace context is set direct through
        # the context variable.... so no reset possible

        base_url = os.environ.get("EVENTIX_URL", "")
        if base_url == "":
            log.error("No EVENTIX_URL set.")
            sys.exit()

        cls.set_base_url(base_url)

        namespace = ""
        if "namespace" in config:
            namespace = config['namespace']

        if namespace == "":
            namespace = os.environ.get("EVENTIX_NAMESPACE", "")

        if namespace == "":
            log.error("No EVENTIX_NAMESPACE set.")
            sys.exit()

        namespace_context_var.set(namespace)



from typing import Any, Dict, Optional

from pydantic_db_backend_common.exceptions import AlreadyExists, NotFound, RevisionConflict
from webexception.webexception import WebException


class TaskNotUnique(WebException):
    status_code = 409

    def __init__(self, uid: str) -> None:
        super().__init__(
            f"Task with uid '{uid}' already exists and task unique is not activated for overwriting.",
            uid=uid
        )


class NoTaskFound(WebException):
    status_code = 204

    def __init__(self, namespace: str) -> None:
        super().__init__(
            f"No task for namespace {namespace}",
            namespace=namespace
        )


class NoTaskFoundForUniqueKey(WebException):
    status_code = 204

    def __init__(self, unique_key: str) -> None:
        super().__init__(
            f"No task for unique_key {unique_key}",
        )
        self.unique_id = unique_key


class TaskNotRegistered(WebException):
    status_code = 404

    def __init__(self, task: str) -> None:
        super().__init__(
            f"Task '{task}' not registered.",
            task=task
        )


class EventixHTTPException(WebException):
    def __init__(
        self,
        status_code: int,
        detail: Any = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(status_code=status_code, detail=detail, headers=headers)


backend_exceptions = [
    AlreadyExists,
    NotFound,
    RevisionConflict
]



import logging

from fastapi import APIRouter
from pydantic_db_backend.backend import Backend

from eventix.functions.task import task_post, task_by_unique_key
from eventix.pydantic.task import TaskModel

log = logging.getLogger(__name__)

router = APIRouter(tags=["task"])


@router.post("/task")
async def route_task_post(task: TaskModel) -> TaskModel:
    return task_post(task)


@router.get("/task/{uid}")
async def route_task_get(uid: str) -> TaskModel:
    # noinspection PyTypeChecker
    return Backend.get_instance(TaskModel, uid)


@router.delete("/task/{uid}")
async def route_task_delete(uid: str) -> None:
    return Backend.delete_uid(TaskModel, uid)


@router.get("/task/by_unique_key/{unique_key}")
async def route_task_(unique_key: str) -> TaskModel:
    return task_by_unique_key(unique_key=unique_key)


@router.delete("/task/by_unique_key/{unique_key}")
async def route_task_(unique_key: str) -> None:
    uid = task_by_unique_key(unique_key=unique_key).uid
    return Backend.delete_uid(TaskModel, uid)


@router.put("/task/{uid}")
async def route_task_put(uid: str, task: TaskModel) -> TaskModel:
    task.uid = uid  # overwrite uid
    # noinspection PyTypeChecker
    return Backend.put_instance(task)
